# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-01-15 02:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0021_auto_20191213_0725"),
    ]

    operations = [
        migrations.AddField(
            model_name="scheduleservice",
            name="multi_callback_enabled",
            field=models.BooleanField(default=False, verbose_name="是否支持多次回调"),
        ),
    ]
