# Generated by Django 2.2.19 on 2021-06-07 09:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0028_auto_20201227_1952"),
    ]

    operations = [
        migrations.AddField(
            model_name="templaterelationship",
            name="always_use_latest",
            field=models.BooleanField(default=False, verbose_name="是否永远使用最新版本"),
        ),
    ]
